open Ast
open Type
open Symbol

let rec ext_type_of_type = function
    IntType(_) -> TInt
  | TypeIdent(_,id) -> TClass(id)
  | FunType(_,ts,t) -> TArrow(List.map ext_type_of_type ts,ext_type_of_type t)

let is_subtype s s' = 
  if s = s' then true
  else 
    match (s,s') with
	TNull,TExt(TClass(_)) -> true
      | TBad,_ -> true
      | _,TBad -> true
      | _ -> false

let badtype_at_pos pos msg =
  Report.error_at_pos pos msg;
  TBad

let no_funtype pos s =
  Report.error_at_pos pos ("function type "^(string_of_int_type s)^" not allowed")

let max_type pos s s' =
  if is_subtype s s' then s'
  else 
    if is_subtype s' s then s
    else badtype_at_pos pos ("incomparable types: "^(string_of_int_type s)^" and "^(string_of_int_type s'))

let comparable pos s s' =
  match s,s' with 
      TExt(TArrow(_)),_ -> no_funtype pos s;false
    | _,TExt(TArrow(_)) -> no_funtype pos s';false 
    | TBad,_ | _,TBad -> true
    | _ -> (max_type pos s s') <> TBad

let rec is_type_wf ((TEnv(bs)) as g) = function
    TInt -> true
  | TArrow(ts,t) -> 
      (List.for_all (is_type_wf g) ts) && (is_type_wf g t)
  | TClass(id) ->
      List.exists (function TBind(id',_,KClass,_) -> id = id' | _ -> false) bs

let add_env pos (TEnv(bs) as g) (TBind(x,_,_,_) as b) =
  if List.exists (function (TBind(id,_,_,_)) -> id = x) bs then
    (Report.error_at_pos pos ("variable "^x^" already defined");g)
  else
    TEnv(bs@[b])

let type_int t = t
let type_ext t = type_int (TExt(t))

let check_subtype pos s t =
  if is_subtype s t then ()
  else 
    Report.error_at_pos pos ("type "^(string_of_int_type s)^" is not subtype of type "^(string_of_int_type t)) 

let symbol_table = Hashtbl.create 10

let clear_symbol_table () = Hashtbl.clear symbol_table

let add_symbol sym value = Hashtbl.add symbol_table sym value
  
let get_symbol_value sym = Hashtbl.find symbol_table sym

let add_symbol_type sym (kind,s) = 
  add_symbol sym (kind,s)

let get_symbol sym =
  try
    Hashtbl.find symbol_table sym
  with
      Not_found -> Report.fail ("trying to retrieve type of unknown symbol "^(string_of_symbol sym))

let dump_symbol_table () =
  Hashtbl.iter (fun s (k,t) ->
		  print_string (string_of_symbol s);print_string " -> ";
		  print_kind k;print_string " : ";
		  print_int_type t;
		  print_newline()) symbol_table

let rec type_of_expression (TEnv(bs) as g) expr = 
  let pos = pos_of_expression expr in
    match expr with
	IntLit(_,_) -> type_ext TInt
      | NullLit(_) -> type_int TNull
      | Ident(_,x) ->
	  (try
	     let TBind(_,sym_x,_,s) = List.find (function TBind(id,_,KVar,s) | TBind(id,_,KFun,s) -> x = id | _ -> false) bs
	     in
	       set_symbol_of_expression sym_x expr;
	       type_int s
	   with
	       Not_found -> badtype_at_pos pos ("variable "^x^" not defined"))
      | UnOp(_,u,e) ->
	  ignore (type_of_expression g e);
	  type_ext TInt
      | If(_,_cond,_then,_else) ->
	  begin
	    let s_cond = type_of_expression g _cond in
	    let s_then = type_of_expression g _then in
	    let s_else = type_of_expression g _else in
	      match s_cond with
		  TExt(TInt) 
		| TBad -> max_type pos s_then s_else
		| _ -> Report.error_at_pos (pos_of_expression _cond) "conditionnal not of type Int";max_type pos s_then s_else
	  end
      | BinOp(_,b,e1,e2) ->
	  begin
	    match (type_of_expression g e1),(type_of_expression g e2) with
		TExt(TInt),TExt(TInt) -> type_ext TInt
	      | s1,s2 when b = Eq || b = Neq -> 
		  ignore (comparable pos s1 s2);
		  type_ext TInt
	      | _ -> type_ext TInt
	  end
      | Apply(_,e,es) ->
	  begin
	    match (type_of_expression g e) with
		TExt(TArrow(ts,t)) ->
		  (try
		     let (es_ts) = List.combine es ts in
		       List.iter (fun (e,t) ->
				    let s_e = type_of_expression g e in
				      check_subtype pos s_e (TExt(t))) es_ts;
		       type_ext t
		   with
		       Invalid_argument(_) -> Report.error_at_pos pos "arity mismatch";type_ext t)
	      | _ -> badtype_at_pos pos "non-functional type"
	  end
      | Block(_,sts,e) ->
	  let g = List.fold_left (fun g s -> type_of_statement g s) g sts in
	    type_of_expression g e
      | New(_,c,es) ->
	  (try
	     let TBind(_,sym_c,_,TGamma(TEnv((ls)))) = List.find (function TBind(id,_,KClass,_) -> id = c | _ -> false) bs
	     in
	       set_symbol_of_expression sym_c expr;
	       let es_ls = List.combine es ls in
		 List.iter (function (e,TBind(l,_,KField,s)) -> 
			      let s_e = type_of_expression g e in
				check_subtype (pos_of_expression e) s_e s) es_ls;
		 type_ext (TClass c)
	   with Not_found -> Report.error_at_pos pos ("class "^c^" not defined");type_ext (TClass c)
	     | Invalid_argument(_) -> Report.error_at_pos pos "arity mismatch";type_ext (TClass c))
      | Select(_,e,l) ->
	    match type_of_expression g e with
		TExt(TClass(c)) ->
		  (try
		     let TBind(_,sym_c,_,TGamma(TEnv(ls))) = List.find (function TBind(id,_,KClass,_) -> id = c | _ -> false) bs
		     in
		       (try
			  let TBind(_,sym_l,_,s) = List.find (function TBind(id,_,KField,_) -> id = l | _ -> false) ls in
			    set_symbol_of_expression sym_l expr;
			    type_int s
			with Not_found -> badtype_at_pos pos ("class "^c^" has no field "^l))
		   with Not_found -> badtype_at_pos pos ("class not defined "^c))
	      | _ -> badtype_at_pos pos "is not a class type"
and type_of_statement (TEnv(bs) as g) st = 
  let pos = pos_of_statement st in
    match st with
	While(_,e,sts) ->
	  begin
	    match type_of_expression g e with 
		TExt(TInt) -> ()
	      | _ -> Report.error_at_pos pos ("conditionnal is not of type int")
	  end;
	  ignore (List.fold_left (fun g st -> type_of_statement g st) g sts);
	  g
      | VarDef(_,(Formal(_,x,t) as f),e) -> 
	  let sym_x = new_symbol x in
	    set_symbol_of_statement sym_x st;
	    let t = ext_type_of_type t in
	      if is_type_wf g t then 
		begin
		  let s_e = type_of_expression g e in
		    check_subtype pos s_e (TExt(t));
		    add_symbol_type sym_x (KVar,(TExt(t)));
		    add_env (pos_of_formal f) g (TBind(x,sym_x,KVar,TExt(t)))
		end
	      else 
		begin
		  Report.error_at_pos pos ("type "^(string_of_ext_type t)^" is not well-formed");
		  g
		end
      | Assign(_,x,e) -> 
	  (try
	     let TBind(_,sym_x,_,s) = List.find (function TBind(id,_,KVar,_) -> id = x | _ -> false) bs in
	       begin
		 set_symbol_of_statement sym_x st;
		 let s_e = type_of_expression g e in
		   check_subtype pos s_e s 
	       end
	   with Not_found -> Report.error_at_pos pos ("undefined variable "^x));
	  g
      | Do(_,e) -> ignore (type_of_expression g e);g

let type_of_declaration (TEnv(bs) as g) d =
  let pos = pos_of_decl d in
    match d with
	ClassDef(_,c,fs) ->
	  let sym_c = new_symbol c in
	    set_symbol_of_decl sym_c d;
	    let g_ = add_env pos g (TBind(c,sym_c,KClass,TGamma(TEnv([])))) in
	    let gc = List.fold_left (fun g (Formal(_,l,t) as f) ->
				       let sym_l = new_symbol l in
					 set_symbol_of_formal sym_l f;
					 let t = ext_type_of_type t in
					   if is_type_wf g_ t then 
					     begin
					       add_symbol_type sym_l (KField,(TExt(t)));
					       add_env (pos_of_formal f) g (TBind(l,sym_l,KField,TExt(t)))
					     end
					   else
					     begin
					       Report.error_at_pos (pos_of_formal f) ("type "^(string_of_ext_type t)^" is not well-formed");
					       g
					     end) (TEnv([])) fs
	    in 
	      add_symbol_type sym_c (KClass,TGamma(gc));
	      add_env pos g (TBind(c,sym_c,KClass,TGamma(gc)))
      | FunDef(_,f,xs,t,e) -> 
	  let sym_f = new_symbol f in
	    set_symbol_of_decl sym_f d;
	    let t = ext_type_of_type t in
	    let t' = TArrow(List.map (function Formal(_,_,t) -> ext_type_of_type t) xs,t) in
	      if is_type_wf g t' then 
		begin
		  add_symbol_type sym_f (KFun,TExt(t'));
		  let g' = add_env pos g (TBind(f,sym_f,KFun,TExt(t'))) in
		  let g'_ = List.fold_left (fun g'_ (Formal(_,x,t) as f) -> 
					      let sym_x = new_symbol x in
						set_symbol_of_formal sym_x f;
						add_symbol_type sym_x (KVar,(TExt (ext_type_of_type t)));
						add_env (pos_of_formal f) g'_ (TBind(x,sym_x,KVar,TExt(ext_type_of_type t)))) g' xs in
		  let s_e = type_of_expression g'_ e in
		    check_subtype pos s_e (TExt(t));
		    g'
		end
	      else 
		begin
		  Report.error_at_pos pos ("type "^(string_of_ext_type t')^" is not well-formed");
		  g
		end

let is_program_wf p =
  match p with
      Program(_,ds,e) ->
	let l0 = [TBind("readInt",readInt_symbol,KFun,TExt(TArrow([],TInt)));
		  TBind("readChar",readChar_symbol,KFun,TExt(TArrow([],TInt)));
		  TBind("printInt",printInt_symbol,KFun,TExt(TArrow([TInt],TClass("Unit"))));
		  TBind("printChar",printChar_symbol,KFun,TExt(TArrow([TInt],TClass("Unit"))))
		 ] in
	  clear_symbol_table ();
	  List.iter (function TBind(_,sym,k,t) -> add_symbol_type sym (k,t)) l0;
	  let g = List.fold_left (fun g d -> type_of_declaration g d) (TEnv(l0)) ds in
	    ignore (type_of_expression g e);
	    !Report.errors = 0

let main filename =
  let in_channel = if filename = "--" then stdin else open_in filename in
  let program = Parser.parse_program in_channel in
    if is_program_wf program then 
      begin
	print_program no_tab program;
	dump_symbol_table ()
      end
    else Report.fail "program not well-typed"
