open Symbol

type position = int * int

type annotation = { position : position; mutable symbol : symbol option }

type program = 
    Program of annotation * (decl list) * expression
and decl = 
    ClassDef of annotation * ident * (formal list) 
  | FunDef of annotation * ident * (formal list) * ttype * expression
and formal = 
    Formal of annotation * ident * ttype 
and ttype = 
    IntType of annotation
  | TypeIdent of annotation * ident
  | FunType of annotation * (ttype list) * ttype
and statement =
    While of annotation * expression * (statement list)
  | VarDef of annotation * formal * expression
  | Assign of annotation * ident * expression
  | Do of annotation * expression
and expression = 
    If of annotation * expression * expression * expression
  | UnOp of annotation * unary * expression
  | BinOp of annotation * binary * expression * expression
  | Ident of annotation * ident
  | NullLit of annotation
  | IntLit of annotation * Int32.t
  | Apply of annotation * expression * (expression list)
  | Block of annotation * (statement list) * expression
  | New of annotation * ident * (expression list)
  | Select of annotation * expression * ident
and unary = 
    Not
  | Neg
and binary =
    Add
  | Sub
  | Mul
  | Div
  | Mod
  | Eq
  | Neq
  | Lt
  | Le
  | Gt
  | Ge
  | Or
and ident = string

let new_annotation pos = { position = pos; symbol = None }

let copy_annotation ann = { position = ann.position; symbol = None }

let pos_of_annotation a = a.position

let pos_of_program = function
    Program(ann,_,_) -> pos_of_annotation ann
and pos_of_decl = function
    ClassDef(ann,_,_) 
  | FunDef(ann,_,_,_,_) -> pos_of_annotation ann
and pos_of_formal = function
    Formal(ann,_,_) -> pos_of_annotation ann
and pos_of_ttype = function
    IntType(ann)
  | TypeIdent(ann,_)
  | FunType(ann,_,_) -> pos_of_annotation ann
and pos_of_statement = function
    While(ann,_,_) 
  | VarDef(ann,_,_)
  | Assign(ann,_,_) 
  | Do(ann,_) -> pos_of_annotation ann
and pos_of_expression = function
    If(ann,_,_,_)
  | UnOp(ann,_,_)
  | BinOp(ann,_,_,_) 
  | Ident(ann,_) 
  | NullLit(ann)
  | IntLit(ann,_)
  | Apply(ann,_,_)
  | Block(ann,_,_)
  | New(ann,_,_)
  | Select(ann,_,_) -> pos_of_annotation ann

let safe_update_of_symbol ann sym =
  match ann.symbol with
      None -> ann.symbol <- Some(sym)
    | Some(sym') ->
	if sym = sym' then ()
	else failwith "symbol already present"

let set_symbol_of_decl sym = function
    ClassDef(ann,_,_) 
  | FunDef(ann,_,_,_,_) -> safe_update_of_symbol ann sym

let set_symbol_of_formal sym = function
    Formal(ann,_,_) -> safe_update_of_symbol ann sym

let set_symbol_of_statement sym = function
    While(ann,_,_) 
  | VarDef(ann,_,_)
  | Assign(ann,_,_) 
  | Do(ann,_) -> safe_update_of_symbol ann sym

let set_symbol_of_expression sym = function
    If(ann,_,_,_)
  | UnOp(ann,_,_)
  | BinOp(ann,_,_,_) 
  | Ident(ann,_) 
  | NullLit(ann)
  | IntLit(ann,_)
  | Apply(ann,_,_)
  | Block(ann,_,_)
  | New(ann,_,_)
  | Select(ann,_,_) -> safe_update_of_symbol ann sym

let no_tab () = ()

let add_tab tab () = print_string "    ";tab ()

let tab_nl tab_f tab_nl =
  let f = ref true in
    function () ->
      if !f then
	begin
	  f := false;
	  tab_f ()
	end
      else tab_nl ()

let print_annotation { position = x; symbol = s } =
  match s with
      None -> ()
    | Some(s) -> () (* ;print_string " /* ";print_string (string_of_symbol s);print_string " */ " *)

let rec print_program tab = function
    Program(_,decls,e) ->
      List.iter (fun d -> print_decl tab d;print_newline()) decls;
      print_expression tab e;
      print_newline()
and print_decl tab = function
    ClassDef(ann,id,fs) ->
      if ((id = "Unit") || (id = "List")) then ()
      else 
	begin
	  tab ();print_string "class ";print_string id;print_string " {\n";
	  List.iter (fun f -> add_tab tab ();print_string "val ";print_formal no_tab f;print_string ";\n") fs;
	  tab ();print_string "}";
	  print_annotation ann;
	  print_newline()
	end
  | FunDef(ann,id,fs,t,e) ->
      tab ();print_string "def ";print_string id;print_string "(";
      begin
	match fs with
	    [] -> ()
	  | [f] -> print_formal no_tab f
	  | f::fs ->
	      print_formal no_tab f;
	      List.iter (fun f -> print_string ", ";print_formal no_tab f) fs
      end;
      print_string ") : ";
      print_type tab t;
      print_string " = ";
      print_expression no_tab e;
      print_string ";";
      print_annotation ann
and print_formal tab = function
    Formal(ann,id,t) -> 
      tab ();print_string id;print_string ": ";print_type no_tab t;print_annotation ann
and print_type tab = function
    IntType(_) -> 
      tab ();print_string "Int"
  | TypeIdent(_,id) -> 
      tab ();print_string id
  | FunType(_,ts,t) -> 
      tab ();print_string "(";
      begin
	match ts with
	    [] -> ()
	  | [t] -> print_type no_tab t
	  | t::ts ->
	      print_type no_tab t;
	      List.iter (fun t -> print_string ", ";print_type no_tab t) ts
      end;
      print_string ") => ";
      print_type no_tab t
and print_statement tab = function
    While(_,e,st) ->
      tab ();print_string "while (";print_expression no_tab e;print_string ") {\n";
      List.iter (fun st -> print_statement (add_tab tab) st;print_newline ()) st;
      tab ();print_string "}"
  | VarDef(ann,f,e) ->
      tab ();print_string "var ";print_formal no_tab f;print_string " = ";print_expression no_tab e;print_string ";";print_annotation ann
  | Assign(ann,id,e) ->
      tab ();print_string "let ";print_string id;print_string " = ";print_expression no_tab e;print_string ";";print_annotation ann
  | Do(_,e) ->
      tab ();print_string "do ";print_expression (tab_nl no_tab (add_tab tab)) e;print_string ";"
and print_expression tab = function 	
    If(_,c,t,e) ->
      tab ();print_string "if (";print_expression no_tab c;print_string ") ";print_expression no_tab t;print_string " else ";print_expression no_tab e
  | UnOp(_,un,e) ->
      tab ();print_unary un;print_string "(";print_expression no_tab e;print_string ")"
  | BinOp(_,bi,e,e') ->
      tab ();print_string "(";print_expression no_tab e;print_binary bi;print_expression no_tab e';print_string ")"
  | Ident(ann,id) -> 
      tab ();print_string id;print_annotation ann
  | NullLit(_) ->
      tab ();print_string "null"
  | IntLit(_,n) ->
      tab ();print_string (Int32.to_string n)
  | Apply(_,e,es) ->
      tab ();print_expression no_tab e;print_params es
  | Block(_,sts,e) ->
      tab ();print_string "{\n";
      List.iter (fun st -> print_statement (add_tab tab) st;print_newline ()) sts;
      print_expression (add_tab tab) e;print_newline();
      tab ();print_string "}"
  | New(ann,id,es) ->
      tab ();print_string "new ";print_string id;print_params es;print_annotation ann
  | Select(ann,e,id) ->
      print_expression tab e;print_string ".";print_string id;print_annotation ann
and print_params es = 
  print_string "(";
  begin
    match es with
	[] -> ()
      | [e] -> print_expression no_tab e
      | e::es ->
	  print_expression no_tab e;
	  List.iter (fun e -> print_string ", ";print_expression no_tab e) es
  end;
  print_string ")"
and print_unary = function
  Not -> print_string "!"
| Neg -> print_string "-"
and print_binary = function
  Add -> print_string "+"
| Sub -> print_string "-"
| Mul -> print_string "*"
| Div -> print_string "/"
| Mod -> print_string "%"
| Eq -> print_string "=="
| Neq -> print_string "!="
| Lt -> print_string "<"
| Le -> print_string "<="
| Gt -> print_string ">"
| Ge -> print_string ">="
| Or -> print_string "|"
