open Ast
open Type
open Symbol
open Risc
open Code

let getCounter = 
  let c = ref 0 in
    function () ->
      let v = !c in incr c;"_"^string_of_int v

type symbol_type =
    GLocalVar of int 
  | GField of int
  | GClass of int 
  | GFun of label 

let symbol_def_table : (symbol,symbol_type) Hashtbl.t = Hashtbl.create 10

let add_symbol_def sym sym_t = 
  Hashtbl.add symbol_def_table sym sym_t

let get_symbol_def sym =
  try
    Hashtbl.find symbol_def_table sym
  with
      Not_found -> Report.fail ("unknown symbol "^(string_of_symbol sym))

let dummy_annotation = new_annotation (-1,-1)

let true_value = IntLit(dummy_annotation,Int32.one)
let false_value = IntLit(dummy_annotation,Int32.zero)

let opt = ref true

let generator program = 
  let program = 
    if !opt && false then 
      Optimiser.optimise program 
    else 
      begin
	if not (Analyzer.is_program_wf program) then 
	  Report.fail "program not well-typed";
	program
      end
  in
  let (get_register,get_free_register,free_register,used_registers) = registers ()
  and (getPC,declare,emit_code,comment_code,print_code) = code () in
  let on_stack = ref 0 in
  let push_on_stack n = 
    on_stack := !on_stack + n
  and pop_from_stack n =
    on_stack := !on_stack - n
  and on_stack () = !on_stack in
  let sp_reg = 30 in
  let rzero = R(0) and rres = R(1) and rsp = R(sp_reg) and rret = R(31) in
  let signed_value n = Signed(function () -> n) in
  let unsigned_value n = Unsigned(function () -> n) in
  let lsh r_source n r_target =
    emit_code (ISh(RLsh,r_target,r_source,BI(signed_value n)))
  in
  let move source r_target =
    emit_code (IArith(RAdd,r_target,rzero,source))
  in
  let move_label label r_target =
    comment_code (label.label^" -> R"^(string_of_register r_target));
    move (AIU(Unsigned(fun () -> getAnchor label))) r_target;
    lsh r_target 2 r_target (* word size = 4 *)
  and bcc cond r_source label =
    comment_code ("branch conditionnally to "^label.label);
    let pc = getPC () in
      emit_code (ITest(cond,r_source,Relative(fun () -> getAnchor label - pc)))
  in
  let bra label =
    comment_code ("branch to "^label.label);
    bcc RBeq rzero label 
  in
  let jmp_register r_jump =
    comment_code ("jump R"^(string_of_register r_jump));
    emit_code (Ret(r_jump))
  in
  let jmp label =
    let r_jump = get_free_register () in
      move_label label r_jump;
      comment_code ("jump to "^label.label);
      jmp_register r_jump;
      free_register r_jump
  and jsr label =
    comment_code ("jump to subroutine "^label.label);
    emit_code (Jsr(Absolute(function () -> getAnchor label)))
  in
  let push reg =
    push_on_stack 1;
    comment_code ("push R"^(string_of_register reg));
    emit_code (IMem(RPsh,reg,rsp,signed_value word_size))
  and pop reg =
    comment_code ("pop R"^(string_of_register reg));
    emit_code (IMem(RPop,reg,rsp,signed_value word_size));
    pop_from_stack 1
  in
  let arith_op op source r_target =
    emit_code (IArith(op,r_target,r_target,source))
  and mem_op op (r_addr,offset) r_data =
    emit_code (IMem(op,r_data,r_addr,signed_value offset))
  in
  let neg r_target =
    emit_code (IArith(RSub,r_target,rzero,AR(r_target)))
  in
  let trap syscall r =
    emit_code (Syscall(r,syscall))
  in
  let save_registers regs = 
    List.iter (fun r -> push r;free_register r) regs
  and restore_registers regs =
    List.iter (fun r -> get_register r;pop r) (List.rev regs)
  in
  let aop_of_bop = function
      Add -> RAdd
    | Sub -> RSub
    | Mul -> RMul
    | Div -> RDiv
    | Mod -> RMod
  and lop_of_bop = function
      Or -> ROr
  and top_of_bop = function
      Eq -> RBeq
    | Neq -> RBne
    | Lt -> RBlt
    | Le -> RBle
    | Gt -> RBgt
    | Ge -> RBge 
  and neg_top = function
      RBeq -> RBne
    | RBne -> RBeq
    | RBlt -> RBge
    | RBge -> RBlt
    | RBle -> RBgt
    | RBgt -> RBle 
  and is_aop = function Add | Sub | Mul | Div | Mod -> true | _ -> false
  and is_top = function Eq | Neq | Lt | Le | Gt | Ge -> true | _ -> false
  in
  let free_stack old_top = 
    let to_free = on_stack () - old_top in
      if to_free > 0 then
	begin
	  arith_op RAdd (AIU(unsigned_value (word_size * to_free))) rsp;
	  pop_from_stack to_free;
	end
      else 
	if to_free < 0 then Report.fail "Strange stack allocation"
  in
  let rec generate_decl = function
      ClassDef(ann,_,_) -> 
	let Some(class_sym) = ann.symbol in
	let (KClass,TGamma(TEnv(fields))) = Analyzer.get_symbol class_sym in
	let all_syms = List.map (function TBind(_,sym,_,_) -> sym) fields in
	let size = List.fold_left (fun i sym -> add_symbol_def sym (GField(i));i+1) 0 all_syms in
	  add_symbol_def class_sym (GClass(size))
    | FunDef(ann,id,args,_,e) ->
	let l__fun = new_label ("def__"^id) in
	  declare l__fun;
	  let Some(fun_sym) = ann.symbol in
	  let (KFun,TExt(TArrow(_in,_out))) = Analyzer.get_symbol fun_sym in
	    add_symbol_def fun_sym (GFun(l__fun));
	    let all_syms = List.map (function Formal(ann,_,_) -> let Some(s) = ann.symbol in s) args in
	    let my_stack = on_stack () in
	    let nb_args = List.fold_left (fun i sym -> 
					    add_symbol_def sym (GLocalVar(on_stack ()));
					    push_on_stack 1;i+1) 0 all_syms in
	      push rret;
	      get_register rres;
	      generate_expression ~tailcall:(Some(my_stack,nb_args)) rres e;
	      free_register rres;
	      pop rret;
	      if on_stack () - my_stack <> nb_args then Report.fail "Wrong stack size";
	      free_stack my_stack;
	      comment_code ("return from subroutine "^l__fun.label);
	      jmp_register rret
  and generate_statement = function
      While(_,_cond,sts) -> 
	let post = getCounter() in
	let l__while = new_label ("__while"^post) in
	let l__wend = new_label ("__wend"^post) in
	  declare l__while;
	  generate_condition true l__wend _cond;
	  let my_stack = on_stack () in
	    List.iter generate_statement sts;
	    free_stack my_stack;
	    bra l__while;
	    declare l__wend
    | VarDef(ann,_,e) -> 
	let Some(var_sym) = ann.symbol in
	let r_e = get_free_register () in
	  generate_expression r_e e;
	  add_symbol_def var_sym (GLocalVar(on_stack()));
	  push r_e;
	  free_register r_e
    | Assign(ann,_,e) ->
	let Some(x_sym) = ann.symbol in
	  begin
	    match get_symbol_def x_sym with
		GLocalVar(pos_on_stack) ->
		  let r_e = get_free_register () in
		    generate_expression r_e e;
		    let offset = word_size * (on_stack () - pos_on_stack - 1) in
		      mem_op RStw (rsp,offset) r_e;
		      free_register r_e
	      | _ -> Report.fail "Strange affectation"
	  end
    | Do(_,e) ->
	let r_e = get_free_register () in
	  generate_expression r_e e;
	  free_register r_e
  and generate_expression ?(tailcall=None) r_target = function
      If(_,_cond,_then,_else) ->
	let post = getCounter() in
	let l__else = new_label ("__else"^post) in
	let l__endif = new_label ("__endif"^post) in
	  generate_condition true l__else _cond;
	  generate_expression r_target ~tailcall:tailcall _then;
	  bra l__endif;
	  declare l__else;
	  generate_expression r_target ~tailcall:tailcall _else;
	  declare l__endif
    | UnOp(_,uop,e) ->
	begin
	  match uop with
	      Not -> 
		generate_expression r_target (If(dummy_annotation,e,false_value,true_value))
	    | Neg ->
		generate_expression r_target e;
		neg r_target
	end
    | BinOp(_,bop,e1,e2) ->
	  if is_aop bop then
	    (* arithmetic operation *)
	    begin
	      generate_expression r_target e1;
	      let r_tmp = get_free_register () in
		generate_expression r_tmp e2;
		arith_op (aop_of_bop bop) (AR(r_tmp)) r_target;
		free_register r_tmp
	    end
	  else 
	    if is_top bop then
	      (* comparison *)
	      generate_expression r_target (If(dummy_annotation,BinOp(dummy_annotation,bop,e1,e2),true_value,false_value))
	    else
	      (* logical operation: OR *)
	      generate_expression r_target (If(dummy_annotation,e1,true_value,If(dummy_annotation,e2,true_value,false_value)));
    | Ident(ann,_) ->
	let Some(x_sym) = ann.symbol in
	  begin
	    match get_symbol_def x_sym with
		GLocalVar(pos_on_stack) ->
		  let offset = word_size * (on_stack () - pos_on_stack - 1) in
		    mem_op RLdw (rsp,offset) r_target
	      | GFun(label) ->
		  move_label label r_target
	      | _ -> Report.fail "Identifier with no symbol"
	  end
    | NullLit(_) -> move (AR(rzero)) r_target
    | IntLit(_,n) -> 
	let q = Int32.to_int (Int32.shift_right n 16) in
	let r = Int32.to_int (Int32.logand n (Int32.of_int ((1 lsl 16)-1))) in
	  comment_code ("load "^(Int32.to_string n)^" to R"^(string_of_register r_target));
	  if q = 0 then 
	    move (AIU(unsigned_value r)) r_target
	  else
	    begin
	      move (AIU(unsigned_value q)) r_target;
	      lsh r_target 16 r_target;
	      if(r <> 0) then
		arith_op RAdd (AIU(unsigned_value r)) r_target;
	    end
    | Apply(_,Ident(ann,_),es) when (let Some(fun_sym) = ann.symbol in match get_symbol_def fun_sym with GFun(_) -> true | _ -> false) && tailcall=None ->
	let Some(fun_sym) = ann.symbol in
	let syscall = List.fold_left (fun o (syssym,syscall) -> if fun_sym = syssym then Some(syscall) else o) None symbol_syscall_assoc in
	let GFun(fun_label) = get_symbol_def fun_sym in
	  begin
	    match syscall with 
		None ->  
		  free_register r_target;
		  let used = used_registers () in
		    save_registers used;
		    let r_e = get_free_register () in
		    let nb_args = List.fold_left (fun i e ->
						    generate_expression r_e e;
						    push r_e;
						    i+1) 0 es in
		      free_register r_e;
		      jsr fun_label;
		      pop_from_stack nb_args;
		      get_register r_target;
		      move (AR(rres)) r_target;
		      restore_registers used
	      | Some(syscall) ->
		  begin
		    match es with 
			[] -> ()
		      | [e] -> generate_expression r_target e
		      | _ -> Report.fail "Unknown system call (wrong arity)"
		  end;
		  comment_code "inlined system call";
		  trap syscall r_target
	  end
    | Apply(ann,e,es) ->
	begin
	  match tailcall with 
	      None -> 
		free_register r_target;
		let used = used_registers () in
		  save_registers used;
		  let r_fun = get_free_register () in
		    generate_expression r_fun e;
		    let r_e = get_free_register () in
		    let nb_args = List.fold_left (fun i e ->
						    generate_expression r_e e;
						    push r_e;
						    i+1) 0 es in
		      free_register r_e;
		      let l__ret = new_label ("return"^getCounter()) in
			move_label l__ret rret;
			comment_code "normal call";
			jmp_register r_fun;
			free_register r_fun;
			declare l__ret;
			pop_from_stack nb_args;
			get_register r_target;
			if rres <> r_target then move (AR(rres)) r_target;
			restore_registers used
	    | Some(stack_f,nb_args_on_stack) -> 
		free_register r_target;
		let used = used_registers () in
		  List.iter (fun r -> free_register r) used;
		  let r_fun = get_free_register () in
		    generate_expression r_fun e;
		    let r_e = get_free_register () in
		    let nb_args = List.fold_left (fun i e ->
						    generate_expression r_e e;
						    push r_e;
						    i+1) 0 es in
		      mem_op RLdw (rsp,word_size*(on_stack()-stack_f-nb_args_on_stack-1)) rret;
		      for i = 0 to nb_args-1 do
			mem_op RLdw (rsp,word_size*(nb_args-i-1)) r_e;
			mem_op RStw (rsp,word_size*(on_stack()-stack_f-i-1)) r_e
		      done;
		      free_register r_e;
		      arith_op RAdd (AIU(unsigned_value (word_size * (on_stack()-stack_f-nb_args)))) rsp;
		      comment_code "tail call";
		      jmp_register r_fun;
		      free_register r_fun;
		      pop_from_stack nb_args;
		      get_register r_target;
		      if rres <> r_target then Report.fail "Wrong tail-call elimination";
		      List.iter (fun r -> get_register r) used
	end;
    | Block(_,sts,e) ->
	let my_stack = on_stack () in
	  List.iter generate_statement sts;
	  generate_expression ~tailcall:tailcall r_target e;
	  free_stack my_stack
    | New(ann,_,es) ->
	let Some(class_sym) = ann.symbol in
	let GClass(size) = get_symbol_def class_sym in
	  if(size > 0) then
	    begin
	      let r_size = get_free_register () in
		move (AIU(unsigned_value (word_size*size))) r_size;
		trap (Sys_gc_alloc(r_size)) r_target;
		free_register r_size;
		push r_target;
		ignore (List.fold_left (fun i e ->
					  generate_expression r_target e;
					  let r_new = get_free_register () in
					    mem_op RLdw (rsp,0) r_new;
					    mem_op RStw (r_new,(word_size * i)) r_target;
					    free_register r_new;
					    i+1) 0 es);
		pop r_target
	    end
	  else
	    trap (Sys_gc_alloc(rzero)) r_target
    | Select(ann,e,_) ->
	let Some(field_sym) = ann.symbol in
	let GField(idx) = get_symbol_def field_sym in
	  generate_expression r_target e;
	  let post = getCounter() in
	  let l__not_null = new_label ("__not_null"^post) in
	    bcc RBne r_target l__not_null;
	    move (AI(signed_value (-1))) r_target;
	    trap Sys_exit r_target;
	    declare l__not_null;
	    mem_op RLdw (r_target,(word_size * idx)) r_target
  and generate_condition iffalse label = function
      BinOp(_,bop,e1,e2) when is_top bop -> 
	let top = top_of_bop bop in
	let top = if iffalse then neg_top top else top in
	let r1 = get_free_register () in
	  generate_expression r1 e1;
	  let r2 = get_free_register () in
	    generate_expression r2 e2;
	    arith_op RCmp (AR(r2)) r1;
	    free_register r2;
	    bcc top r1 label;
	    free_register r1
    | BinOp(_,Or,e1,e2) ->
	let l__short = new_label ("__short"^getCounter()) in
	  generate_condition (not iffalse) l__short e1;
	  generate_condition iffalse label e2;
	  declare l__short;
    | UnOp(_,Not,e) -> 
	generate_condition (not iffalse) label e
    | e ->
	let top = RBne in
	let top = if iffalse then neg_top top else top in
	let r_target = get_free_register () in
	  generate_expression r_target e;
	  bcc top r_target label;
	  free_register r_target
  in
  let generate_wrapper (fun_sym,syscall) =
    let l__fun = new_label ("wrapper__"^(string_of_symbol fun_sym)) in
      declare l__fun;
      let (KFun,TExt(TArrow(_in,_out))) = Analyzer.get_symbol fun_sym in
      let arg_opt = 
	match _in with
	    [] -> false
	  | [x] -> true
	  | _ -> Report.fail "Wrong system call"
      in
	add_symbol_def fun_sym (GFun(l__fun));
	get_register rres;
	if arg_opt then
	  begin
	    push_on_stack 1;
	    pop rres
	  end;
	trap syscall rres;
	free_register rres;
	comment_code ("return from wrapper "^l__fun.label);
	jmp_register rret
  in
  let generate_program = function
      Program(ann,decls,e) ->
	let l__start = new_label "__start" in
	let l__end = new_label "__end" in
	  declare l__start;
	  trap Sys_get_total_mem_size rsp;
	  lsh rsp (-2) (R(1)); (* size in words which is also the begin of heap *)
	  lsh (R(1)) (-1) (R(2)); (* size in words to allocate *)
	  move (AIU(unsigned_value sp_reg)) (R(3));
	  lsh (R(3)) 27 (R(3));
	  arith_op RAdd (AR(R(3))) (R(2));
	  trap (Sys_gc_init(R(2))) (R(1));
	  let l__main = new_label "__main" in
	    jmp l__main;
	    List.iter generate_wrapper symbol_syscall_assoc;
	    List.iter generate_decl decls;
	    declare l__main;
	    let r_e = get_free_register () in
	      generate_expression r_e e;
	      free_register r_e;
	      trap Sys_exit rzero;
	      declare l__end;
	      (match used_registers () with [] -> () | _ -> Report.error "Some registers were not freed");
	      if on_stack () <> 0 then Report.fail "Invalid stack";
	      print_code ()
  in
    generate_program program

let main filename =
  let in_channel = if filename = "--" then stdin else open_in filename in
  let program = Parser.parse_program in_channel in
    generator program
