open Ast
open Symbol

let false_value = Int32.zero
let true_value = Int32.one

let is_equal n m = (Int32.compare n m) = 0
let is_lt n m = (Int32.compare n m) < 0

let is_false_value n = is_equal false_value n 

let uop_of_value n = function
    Not -> if is_false_value n then true_value else false_value
  | Neg -> Int32.neg n

let bop_of_value n m = function
    Add -> Int32.add n m
  | Sub -> Int32.sub n m
  | Mul -> Int32.mul n m
  | Div -> if is_equal m (Int32.zero) then Report.fail "Division by zero";Int32.div n m
  | Mod -> if is_equal m (Int32.zero) then Report.fail "Division by zero";Int32.rem n m
  | Eq -> if is_equal n m then true_value else false_value
  | Neq -> if is_equal n m then false_value else true_value
  | Lt -> if is_lt n m then true_value else false_value
  | Le -> if (is_lt n m) || (is_equal n m) then true_value else false_value
  | Gt -> if (is_lt m n) then true_value else false_value
  | Ge -> if (is_lt m n) || (is_equal n m) then true_value else false_value

type var_type = Variable | Constant of expression
type sym_info = { mutable used : bool; mutable vtype : var_type option }

let symbol_info = Hashtbl.create 10

let symbol_info_clear () = Hashtbl.clear symbol_info

let symbol_used sym =
  try
    let info = Hashtbl.find symbol_info sym
    in info.used <- true
  with
      Not_found -> Hashtbl.add symbol_info sym { used = true; vtype = None }

let symbol_constant sym n =
  try
    let info = Hashtbl.find symbol_info sym
    in 
      info.vtype <- Some(Constant(n))
  with
      Not_found -> Hashtbl.add symbol_info sym { used = false; vtype = Some(Constant(n)) }
  
let symbol_variable sym = 
  try
    let info = Hashtbl.find symbol_info sym
    in 
      info.vtype <- Some(Variable)
  with
      Not_found -> Hashtbl.add symbol_info sym { used = false; vtype = Some(Variable) }

let get_symbol_info sym =
  try
    Hashtbl.find symbol_info sym
  with Not_found -> { used = false; vtype = None }

let defs = Hashtbl.create 10

let defs_clear () = Hashtbl.clear defs

let defs_add sym def = Hashtbl.add defs sym def

let is_defined sym = Hashtbl.mem defs sym

let get_def sym = Hashtbl.find defs sym

let dummy_annotation = new_annotation (-1,-1)

let new_symbol = 
  let count = ref 0 in
    function id ->
      let s = string_of_int (!count) in
	incr count;
	let id' = id^"__"^s in
	  new_symbol (id^"__"^s)

let phase_zero program =
  defs_clear ();
  let rec fill_program = function
      Program(_,ds,e) -> List.iter fill_decl ds
  and fill_decl d =
    match d with
      ClassDef(ann,_,_) 
    | FunDef(ann,_,_,_,_) ->
	let Some(sym) = ann.symbol in
	  defs_add sym d
  in
    fill_program program

let rec duplicate_expression (assoc,e) = 
  match e with
      If(ann,_cond,_then,_else) -> 
	let (assoc,_cond) = duplicate_expression (assoc,_cond) in
	let (assoc,_then) = duplicate_expression (assoc,_then) in
	let (assoc,_else) = duplicate_expression (assoc,_else) in
	  assoc,If(ann,_cond,_then,_else)
    | Block(ann,sts,e) -> 
	let (assoc,sts) = List.fold_left (fun (assoc,sts) st ->
					    let (assoc,st) = duplicate_statement (assoc,st) in
					      (assoc,st::sts)) (assoc,[]) sts 
	in
	let sts = List.rev sts in
	let (assoc,e) = duplicate_expression (assoc,e) in
	  assoc,Block(ann,sts,e)
    | UnOp(ann,uop,e) -> 
	let (assoc,e) = duplicate_expression (assoc,e) in
	  assoc,UnOp(ann,uop,e)
    | Apply(ann,e,es) -> 
	let (assoc,e) = duplicate_expression (assoc,e) in
	let (assoc,es) = List.fold_left (fun (assoc,es) e ->
					   let (assoc,e) = duplicate_expression (assoc,e) in
					     (assoc,e::es)) (assoc,[]) es
	in
	let es = List.rev es in
	  assoc,Apply(ann,e,es)
    | New(ann,id,es) -> 
	let (assoc,es) = List.fold_left (fun (assoc,es) e ->
					   let (assoc,e) = duplicate_expression (assoc,e) in
					     (assoc,e::es)) (assoc,[]) es
	in
	let es = List.rev es in
	  assoc,New(ann,id,es)
    | Select(ann,e,id) ->
	let (assoc,e) = duplicate_expression (assoc,e) in
	  assoc,Select(ann,e,id)
    | BinOp(ann,bop,e1,e2) -> 
	let (assoc,e1) = duplicate_expression (assoc,e1) in
	let (assoc,e2) = duplicate_expression (assoc,e2) in
	  assoc,BinOp(ann,bop,e1,e2)
    | Ident(ann,id) -> 
	begin
	  let Some(sym) = ann.symbol in
	    try
	      let sym' = List.assoc sym assoc in
	      let ann = copy_annotation ann in
		ann.symbol <- Some(sym');
		symbol_used sym';
		assoc,Ident(ann,id)
	    with
		Not_found -> assoc,e
	end
    | e -> assoc,e
and duplicate_statement (assoc,st) = 
  match st with
      While(ann,e,sts) -> 
	let (assoc,e) = duplicate_expression (assoc,e) in
	let (assoc,sts) = List.fold_left (fun (assoc,sts) st -> 
					    let assoc,st = duplicate_statement (assoc,st) in
					      assoc,st::sts) (assoc,[]) sts
	in
	let sts = List.rev sts in
	  assoc,While(ann,e,sts)
    | VarDef(ann,Formal(annf,id,t),e) -> 
	let Some(sym) = ann.symbol in
	let sym' = new_symbol id in
	let annf = copy_annotation annf in
	  annf.symbol <- Some(sym');
	  let ann = copy_annotation ann in
	    ann.symbol <- Some(sym');
	    let (assoc,e) = duplicate_expression (assoc,e) in
	      ((sym,sym')::assoc),VarDef(ann,Formal(annf,id,t),e)
    | Assign(ann,id,e) -> 
	let (assoc,e) = duplicate_expression (assoc,e) in
	let Some(sym) = ann.symbol in
	let sym' = List.assoc sym assoc in (* normalement a t dupliqu *)
	  symbol_used sym';
	  let ann = copy_annotation ann in
	    ann.symbol <- Some(sym');
	    assoc,Assign(ann,id,e)
    | Do(ann,e) -> 
	let (assoc,e) = duplicate_expression (assoc,e) in
	  assoc,Do(ann,e)

let phase_one program = 
  symbol_info_clear ();
  let rec optimise_expression = function
      If(ann,_cond,_then,_else) ->
	begin
	  match optimise_expression _cond with
	      IntLit(_,n) ->
		if is_false_value n then
		  optimise_expression _else
		else
		  optimise_expression _then
	    | _cond -> If(ann,_cond,optimise_expression _then,optimise_expression _else)
	end
    | Block(ann,sts,e) ->
	begin
	  let sts = List.fold_left (fun sts st -> sts@(optimise_statement st)) [] sts in
	  let e = optimise_expression e in
	    match sts with
		[] -> e
	      | sts -> 
		  begin
		    match e with
			Block(ann',sts',e) -> Block(ann,sts@sts',e)
		      | _ -> Block(ann,sts,e)
		  end
	end
    | UnOp(ann,uop,e) ->
	begin
	  match optimise_expression e with
	      IntLit(_,n) -> IntLit(ann,uop_of_value n uop)
	    | e -> UnOp(ann,uop,e)
	end
    | Apply(ann,e,es) -> 
	begin
	  let e = optimise_expression e in
	  let es = List.map optimise_expression es in
	    match e with
		Ident(annf,f) ->
		  let Some(fun_sym) = annf.symbol in
		    if is_defined fun_sym then
		      if List.exists (fun e -> match e with IntLit(_) | NullLit(_) -> true | _ -> false) es then
			let FunDef(ann_fun,id,fs,t,e) = get_def fun_sym in
			let fs_es = List.combine fs es in
			let assoc_ds = List.map (fun (Formal(ann,x,t),e) ->
						let Some(sym) = ann.symbol in
						let sym' = new_symbol x in
						let ann = copy_annotation ann in
						  ann.symbol <- Some(sym');
						  (sym,sym'),(VarDef(ann,Formal(ann,x,t),e))) fs_es in
			let assoc = List.map (fun (x,_) -> x) assoc_ds
			and ds = List.map (fun (_,y) -> y) assoc_ds in
			let _,e = duplicate_expression (assoc,e) in
			  Block(ann,ds,e)
		      else Apply(ann,e,es)
		    else Apply(ann,e,es)
	      | _ -> Apply(ann,e,es)
	end
    | New(ann,id,es) -> 
	let Some(sym) = ann.symbol in symbol_used sym;
	New(ann,id,List.map optimise_expression es)
    | Select(ann,e,id) -> 
	begin
	  match optimise_expression e with
	      New(ann_new,id,es) -> 
		let Some(sel_sym) = ann.symbol 
		and Some(new_sym) = ann_new.symbol in
		let ClassDef(_,id,fs) = get_def new_sym in
		let _,Some(idx,id,t) = List.fold_left (fun (i,o) (Formal(annf,id,t)) ->
							 let Some(l_sym) = annf.symbol in
							   if l_sym = sel_sym then (i,Some(i,id,t))
							   else (i+1,o)) (0,None) fs
		in
		let sym = new_symbol id in
		let ann = new_annotation (-1,-1) in ann.symbol <- Some(sym);
		  let _,sts = List.fold_left (fun (i,sts) e -> 
						let st = 
						  if i = idx then
						    VarDef(ann,Formal(dummy_annotation,id,t),e)
						  else Do(dummy_annotation,e)
						in (i+1,st::sts)) (0,[]) es 
		  in
		  let sts = List.rev sts in
		    symbol_used sym;
		    Block(dummy_annotation,sts,Ident(ann,id))
	    | e -> Select(ann,optimise_expression e,id)
	end
    | BinOp(ann,bop,e1,e2) ->
	begin
	  let e1 = optimise_expression e1 
	  and e2 = optimise_expression e2 in
	    match e1,e2,bop with
		(IntLit(_,n),IntLit(_,m),Or) ->
		  if is_false_value n then 
		    if is_false_value m then IntLit(ann,false_value)
		    else IntLit(ann,true_value)
		  else IntLit(ann,true_value)
	      | (IntLit(_,n),e2,Or) ->
		  if is_false_value n then e2
		  else IntLit(ann,true_value)
	      | (_,IntLit(ann_m,m),Or) ->
		  if is_false_value m then e1
		  else BinOp(ann,Or,e1,IntLit(ann_m,true_value))
	      | (IntLit(_,n),IntLit(_,m),_) -> IntLit(ann,bop_of_value n m bop)
	      | _ -> BinOp(ann,bop,e1,e2)
	end
    | Ident(ann,id) -> 
	let Some(sym) = ann.symbol in symbol_used sym;
	  Ident(ann,id)
    | e -> e
  and optimise_statement = function
      While(ann,_cond,sts) ->
	begin 
	  let sts = List.fold_left (fun sts st -> sts@(optimise_statement st)) [] sts in
	    match optimise_expression _cond with
		IntLit(ann_n,n) ->
		  if is_false_value n then []
		  else [While(ann,IntLit(ann_n,Int32.one),sts)]
	      | _cond -> [While(ann,_cond,sts)]
	end
    | VarDef(ann,f,e) -> 
	begin
	  let e = optimise_expression e in
	  let Some(sym) = ann.symbol in
	    begin
	      match e with
		  IntLit(_) 
		| NullLit(_) -> symbol_constant sym e
		| e -> symbol_variable sym
	    end;
	    [VarDef(ann,f,e)]
	end
    | Assign(ann,x,e) -> 
	begin
	  let Some(sym) = ann.symbol in symbol_variable sym;
	  [Assign(ann,x,optimise_expression e)]
	end
    | Do(ann,e) -> 
	begin
	  match optimise_expression e with
	      IntLit(_)
	    | NullLit(_) -> []
	    | New(_,_,es) -> List.map (function e -> Do(ann,e)) es
	    | Select(_,e,_) -> [Do(ann,e)]
	    | Ident(_,_) -> []
	    | Block(_,sts,e) -> sts@[Do(ann,e)] 
	    | e -> [Do(ann,e)]
	end
  and optimise_decl = function
      ClassDef(ann,c,fs) -> ClassDef(ann,c,fs)
    | FunDef(ann,f,fs,t,e) -> FunDef(ann,f,fs,t,optimise_expression e)
  and optimise_program = function
      Program(ann,ds,e) -> Program(ann,List.map optimise_decl ds,optimise_expression e)
  in
    optimise_program program

let phase_two program =
  let rec optimise_expression = function
      If(ann,_cond,_then,_else) -> If(ann,optimise_expression _cond,optimise_expression _then,optimise_expression _else)
    | Block(ann,sts,e) -> Block(ann,List.map optimise_statement sts,optimise_expression e)
    | UnOp(ann,uop,e) -> UnOp(ann,uop,optimise_expression e)
    | BinOp(ann,bop,e1,e2) -> BinOp(ann,bop,optimise_expression e1,optimise_expression e2)
    | Apply(ann,e,es) -> Apply(ann,optimise_expression e,List.map optimise_expression es)
    | New(ann,id,es) -> New(ann,id,List.map optimise_expression es)
    | Select(ann,e,id) -> Select(ann,optimise_expression e,id)
    | Ident(ann,id) -> 
	begin
	  let Some(sym) = ann.symbol in
	  let info = get_symbol_info sym in
	    match info.vtype with
		Some(Constant(e)) -> e
	      | _ -> Ident(ann,id)
	end
    | IntLit(ann,id) -> IntLit(ann,id)
    | NullLit(ann) -> NullLit(ann)
  and optimise_statement = function
      While(ann,_cond,sts) -> While(ann,optimise_expression _cond,List.map optimise_statement sts)
    | VarDef(ann,f,e) -> 
	begin
	  let Some(sym) = ann.symbol in
	  let info = get_symbol_info sym in
	  let e = optimise_expression e in
	    if info.used then
	      begin
		match info.vtype with
		    Some(Constant(e)) -> Do(ann,e)
		  | _ -> VarDef(ann,f,e)
	      end
	    else Do(ann,e)
	end
    | Assign(ann,x,e) -> Assign(ann,x,optimise_expression e) 
    | Do(ann,e) -> Do(ann,optimise_expression e)
  and optimise_decl = function
      ClassDef(ann,c,fs) -> 
	begin
	  let Some(sym) = ann.symbol in
	  let info = get_symbol_info sym in
	    if info.used then Some(ClassDef(ann,c,fs))
	    else None
	end
    | FunDef(ann,f,fs,t,e) -> 
	begin
	  let Some(sym) = ann.symbol in
	  let info = get_symbol_info sym in
	    if info.used then Some(FunDef(ann,f,fs,t,optimise_expression e))
	    else None
	end
  and optimise_program = function
      Program(ann,ds,e) ->
	let ds = List.map (function Some(d) -> d)
		   (List.filter (function Some(_) -> true | _ -> false)
		      (List.map optimise_decl ds))
	in
	  Program(ann,ds,optimise_expression e)
  in
    optimise_program program

let optimise program = 
  if not (Analyzer.is_program_wf program) then 
    Report.fail "program not well-typed";
  let rec iterate program =
    let program' = phase_one (phase_two program) in
      if program = program' then program
      else iterate program'
  in
    phase_zero program;
    iterate (phase_one program) 

let main filename =
  let in_channel = if filename = "--" then stdin else open_in filename in
  let program = Parser.parse_program in_channel in
    print_program no_tab (optimise program)
