let word_size = 4

type register = R of int

type ic = Signed of (unit -> int)
type uc = Unsigned of (unit -> int)
type oc = Relative of (unit -> int)
type lc = Absolute of (unit -> int)

type a_riiu = AR of register | AI of ic | AIU of uc

type b_ri = BR of register | BI of ic

type arith_op = RAdd | RSub | RMul | RDiv | RCmp | RMod
type logical_op = RAnd | ROr | RXor | RBic
type logical_sh = RLsh | RAsh
type mem_op = RLdw | RLdb | RStw | RStb | RPop | RPsh
type test_op = RBeq | RBne | RBlt | RBge | RBgt | RBle

type syscall = 
    Sys_io_rd_chr
  | Sys_io_rd_int
  | Sys_io_wr_chr
  | Sys_io_wr_int
  | Sys_gc_init of register
  | Sys_gc_alloc of register
  | Sys_get_total_mem_size
  | Sys_exit

type instruction = 
    IArith of arith_op * register * register * a_riiu
  | ILog of logical_op * register * register * a_riiu
  | ISh of logical_sh * register * register * b_ri
  | IMem of mem_op * register * register * ic
  | ITest of test_op * register * oc
  | Chk of register * a_riiu
  | Bsr of oc
  | Jsr of lc
  | Ret of register
  | Break
  | Syscall of register * syscall

let string_of_ic (Signed(f)) = string_of_int (f ())
let string_of_uc (Unsigned(f)) = string_of_int (f ())
let string_of_oc (Relative(f)) = string_of_int (f ())
let string_of_lc (Absolute(f)) = string_of_int (f ())
let string_of_register (R(n)) = string_of_int n

let string_of_a a =
  match a with
      AR(reg) -> string_of_register reg
    | AI(i) -> string_of_ic i
    | AIU(u) -> string_of_uc u

let string_of_b b =
  match b with
      BR(reg) -> string_of_register reg
    | BI(i) -> string_of_ic i 

let postfix_of_a = function
    AR(_) -> ""
  | AI(_) -> "I"
  | AIU(_) -> "IU"

let postfix_of_b = function
    BR(_) -> ""
  | BI(_) -> "I"

let int_of_syscall = function
  | Sys_io_rd_chr -> 1
  | Sys_io_rd_int -> 2
  | Sys_io_wr_chr -> 6
  | Sys_io_wr_int -> 7
  | Sys_gc_init(_) -> 11
  | Sys_gc_alloc(_) -> 12
  | Sys_get_total_mem_size -> 13
  | Sys_exit -> 19

let arg_of_syscall = function
  | Sys_io_rd_chr 
  | Sys_io_rd_int 
  | Sys_io_wr_chr 
  | Sys_io_wr_int
  | Sys_get_total_mem_size 
  | Sys_exit -> R(0)
  | Sys_gc_init(reg) 
  | Sys_gc_alloc(reg) -> reg

let string_of_arith = function
    RAdd -> "ADD"
  | RSub -> "SUB"
  | RMul -> "MUL"
  | RDiv -> "DIV"
  | RCmp -> "CMP"
  | RMod -> "MOD"

let string_of_log = function
    RAnd -> "AND"
  | ROr -> "OR"
  | RXor -> "XOR"
  | RBic -> "BIC"

let string_of_sh = function
    RLsh -> "LSH"
  | RAsh -> "ASH"

let string_of_mem = function
    RLdw -> "LDW"
  | RLdb -> "LDB"
  | RStw -> "STW"
  | RStb -> "STB"
  | RPop -> "POP"
  | RPsh -> "PSH"

let string_of_test = function
    RBeq -> "BEQ"
  | RBne -> "BNE"
  | RBlt -> "BLT"
  | RBge -> "BGE"
  | RBgt -> "BGT"
  | RBle -> "BLE"

let mnemo_of_instruction = function
    IArith(op,_,_,arg) -> (string_of_arith op)^(postfix_of_a arg)
  | ILog(op,_,_,arg) -> (string_of_log op)^(postfix_of_a arg)
  | ISh(op,_,_,arg) -> (string_of_sh op)^(postfix_of_b arg)
  | IMem(op,_,_,_) -> (string_of_mem op)
  | ITest(op,_,_) -> (string_of_test op)
  | Chk(_,arg) -> "CHK"^(postfix_of_a arg)
  | Bsr(_) -> "BSR"
  | Jsr(_) -> "JSR"
  | Ret(_) -> "RET"
  | Break -> "BREAK"
  | Syscall(_) -> "SYSCALL"

let string_of_instruction instr = 
  let sep = " " in
    (mnemo_of_instruction instr)^sep^(
      match instr with
	  IArith(_,a,b,c) 
	| ILog(_,a,b,c) -> (string_of_register a)^sep^(string_of_register b)^sep^(string_of_a c)
	| ISh(_,a,b,c) -> (string_of_register a)^sep^(string_of_register b)^sep^(string_of_b c)
	| IMem(_,a,b,i) -> (string_of_register a)^sep^(string_of_register b)^sep^(string_of_ic i)
	| ITest(_,a,o) -> (string_of_register a)^sep^(string_of_oc o)
	| Chk(a,c) -> (string_of_register a)^sep^(string_of_a c)
	| Bsr(o) -> (string_of_oc o)
	| Jsr(l) ->(string_of_lc l)
	| Ret(c) -> (string_of_register c)
	| Break -> ""
	| Syscall(a,sysc) -> (string_of_register a)^sep^(string_of_register (arg_of_syscall sysc))^sep^(string_of_int (int_of_syscall sysc))
    )
